@@ -3,7 +3,7 @@
    Lexical scanner for dhcpd config file... */
 
 /*
- * Copyright (c) 2004-2009 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2007 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -22,31 +22,27 @@
  *   950 Charter Street
  *   Redwood City, CA 94063
  *   <info@isc.org>
- *   https://www.isc.org/
+ *   http://www.isc.org/
  *
  * This software has been written for Internet Systems Consortium
  * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
  * To learn more about Internet Systems Consortium, see
- * ``https://www.isc.org/''.  To learn more about Vixie Enterprises,
+ * ``http://www.isc.org/''.  To learn more about Vixie Enterprises,
  * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
  * ``http://www.nominum.com''.
  */
 
-#ifndef lint
-static char copyright[] =
-"$Id: conflex.c,v 1.105.8.7 2009/09/01 20:32:27 dhankins Exp $ Copyright (c) 2004-2009 Internet Systems Consortium.  All rights reserved.\n";
-#endif /* not lint */
-
 #include "dhcpd.h"
 #include <ctype.h>
 
 static int get_char PROTO ((struct parse *));
-static enum dhcp_token get_token PROTO ((struct parse *));
+static void unget_char(struct parse *, int);
 static void skip_to_eol PROTO ((struct parse *));
+static enum dhcp_token read_whitespace(int c, struct parse *cfile);
 static enum dhcp_token read_string PROTO ((struct parse *));
 static enum dhcp_token read_number PROTO ((int, struct parse *));
 static enum dhcp_token read_num_or_name PROTO ((int, struct parse *));
-static enum dhcp_token intern PROTO ((unsigned char *, enum dhcp_token));
+static enum dhcp_token intern PROTO ((char *, enum dhcp_token));
 
 isc_result_t new_parse (cfile, file, inbuf, buflen, name, eolp)
 	struct parse **cfile;
@@ -58,34 +54,45 @@ isc_result_t new_parse (cfile, file, inbuf, buflen, name, eolp)
 {
 	struct parse *tmp;
 
-	tmp = dmalloc (sizeof (struct parse), MDL);
-	if (!tmp)
+	tmp = dmalloc(sizeof(struct parse), MDL);
+	if (tmp == NULL) {
 		return ISC_R_NOMEMORY;
-	memset (tmp, 0, sizeof *tmp);
+	}
 
-	tmp -> token = 0;
-	tmp -> tlname = name;
-	tmp -> lpos = tmp -> line = 1;
-	tmp -> cur_line = tmp -> line1;
-	tmp -> prev_line = tmp -> line2;
-	tmp -> token_line = tmp -> cur_line;
-	tmp -> cur_line [0] = tmp -> prev_line [0] = 0;
-	tmp -> warnings_occurred = 0;
-	tmp -> file = file;
-	tmp -> eol_token = eolp;
+	/*
+	 * We don't need to initialize things to zero here, since 
+	 * dmalloc() returns memory that is set to zero.
+	 */
+	/* tmp->token = 0; */ 
+	/* tmp->warnings_occurred = 0; */
+	/* tmp->bufix = 0; */
+	/* tmp->saved_state = NULL; */
+	tmp->tlname = name;
+	tmp->lpos = tmp -> line = 1;
+	tmp->cur_line = tmp->line1;
+	tmp->prev_line = tmp->line2;
+	tmp->token_line = tmp->cur_line;
+	tmp->cur_line[0] = tmp->prev_line[0] = 0;
+	tmp->file = file;
+	tmp->eol_token = eolp;
 
-	tmp -> bufix = 0;
-	tmp -> buflen = buflen;
-	if (inbuf) {
-		tmp -> bufsiz = 0;
-		tmp -> inbuf = inbuf;
+	if (inbuf != NULL) {
+		tmp->inbuf = inbuf;
+		tmp->buflen = buflen;
+		tmp->bufsiz = 0;
 	} else {
-		tmp -> inbuf = dmalloc (8192, MDL);
-		if (!tmp -> inbuf) {
-			dfree (tmp, MDL);
-			return ISC_R_NOMEMORY;
+		struct stat sb;
+
+		if (fstat(file, &sb) < 0)
+			return ISC_R_IOERROR;
+
+		tmp->bufsiz = tmp->buflen = (size_t)sb.st_size;
+		tmp->inbuf = mmap(NULL, tmp->bufsiz, PROT_READ, MAP_SHARED,
+				  file, 0);
+
+		if (tmp->inbuf == MAP_FAILED) {
+			return ISC_R_IOERROR;
 		}
-		tmp -> bufsiz = 8192;
 	}
 
 	*cfile = tmp;
@@ -96,96 +103,156 @@ isc_result_t end_parse (cfile)
 	struct parse **cfile;
 {
 	/* "Memory" config files have no file. */
-	if ((*cfile)->file != -1)
+	if ((*cfile)->file != -1) {
+		munmap((*cfile)->inbuf, (*cfile)->bufsiz);
 		close((*cfile)->file);
+	}
 
-	if ((*cfile)->bufsiz)
-		dfree((*cfile)->inbuf, MDL);
+	if ((*cfile)->saved_state != NULL) {
+		dfree((*cfile)->saved_state, MDL);
+	}
+		
 	dfree(*cfile, MDL);
 	*cfile = NULL;
 	return ISC_R_SUCCESS;
 }
 
+/*
+ * Save the current state of the parser.
+ *
+ * Only one state may be saved. Any previous saved state is
+ * lost.
+ */
+isc_result_t
+save_parse_state(struct parse *cfile) {
+	/*
+	 * Free any previous saved state.
+	 */
+	if (cfile->saved_state != NULL) {
+		dfree(cfile->saved_state, MDL);
+	}
+
+	/*
+	 * Save our current state.
+	 */
+	cfile->saved_state = dmalloc(sizeof(struct parse), MDL);
+	if (cfile->saved_state == NULL) {
+		return ISC_R_NOMEMORY;
+	}
+	memcpy(cfile->saved_state, cfile, sizeof(*cfile));
+	return ISC_R_SUCCESS;
+}
+
+/*
+ * Return the parser to the previous saved state.
+ *
+ * You must call save_parse_state() before calling 
+ * restore_parse_state(), but you can call restore_parse_state() any
+ * number of times after that.
+ */
+isc_result_t
+restore_parse_state(struct parse *cfile) {
+	struct parse *saved_state;
+
+	if (cfile->saved_state == NULL) {
+		return ISC_R_NOTYET;
+	}
+
+	saved_state = cfile->saved_state;
+	memcpy(cfile, saved_state, sizeof(*cfile));
+	cfile->saved_state = saved_state;
+	return ISC_R_SUCCESS;
+}
+
 static int get_char (cfile)
 	struct parse *cfile;
 {
 	/* My kingdom for WITH... */
 	int c;
 
-	if (cfile -> bufix == cfile -> buflen) {
-		if (cfile -> file != -1) {
-			cfile -> buflen =
-				read (cfile -> file,
-				      cfile -> inbuf, cfile -> bufsiz);
-			if (cfile -> buflen == 0) {
-				c = EOF;
-				cfile -> bufix = 0;
-			} else if (cfile -> buflen < 0) {
-				c = EOF;
-				cfile -> bufix = cfile -> buflen = 0;
-			} else {
-				c = cfile -> inbuf [0];
-				cfile -> bufix = 1;
-			}
-		} else
-			c = EOF;
-	} else {
-		c = cfile -> inbuf [cfile -> bufix];
-		cfile -> bufix++;
+	if (cfile->bufix == cfile->buflen)
+		c = EOF;
+	else {
+		c = cfile->inbuf [cfile->bufix];
+		cfile->bufix++;
 	}
 
-	if (!cfile -> ugflag) {
+	if (!cfile->ugflag) {
 		if (c == EOL) {
-			if (cfile -> cur_line == cfile -> line1) {	
-				cfile -> cur_line = cfile -> line2;
-				cfile -> prev_line = cfile -> line1;
+			if (cfile->cur_line == cfile->line1) {	
+				cfile->cur_line = cfile->line2;
+				cfile->prev_line = cfile->line1;
 			} else {
-				cfile -> cur_line = cfile -> line1;
-				cfile -> prev_line = cfile -> line2;
+				cfile->cur_line = cfile->line1;
+				cfile->prev_line = cfile->line2;
 			}
-			cfile -> line++;
-			cfile -> lpos = 1;
-			cfile -> cur_line [0] = 0;
+			cfile->line++;
+			cfile->lpos = 1;
+			cfile->cur_line [0] = 0;
 		} else if (c != EOF) {
-			if (cfile -> lpos <= 80) {
-				cfile -> cur_line [cfile -> lpos - 1] = c;
-				cfile -> cur_line [cfile -> lpos] = 0;
+			if (cfile->lpos <= 80) {
+				cfile->cur_line [cfile->lpos - 1] = c;
+				cfile->cur_line [cfile->lpos] = 0;
 			}
-			cfile -> lpos++;
+			cfile->lpos++;
 		}
 	} else
-		cfile -> ugflag = 0;
+		cfile->ugflag = 0;
 	return c;		
 }
 
-static enum dhcp_token get_token (cfile)
-	struct parse *cfile;
-{
+/*
+ * Return a character to our input buffer.
+ */
+static void
+unget_char(struct parse *cfile, int c) {
+	if (c != EOF) {
+		cfile->bufix--;
+		cfile->ugflag = 1;	/* do not put characters into
+					   our error buffer on the next
+					   call to get_char() */
+	}
+}
+
+/*
+ * GENERAL NOTE ABOUT TOKENS
+ *
+ * We normally only want non-whitespace tokens. There are some 
+ * circumstances where we *do* want to see whitespace (for example
+ * when parsing IPv6 addresses).
+ *
+ * Generally we use the next_token() function to read tokens. This 
+ * in turn calls get_next_token, which does *not* return tokens for
+ * whitespace. Rather, it skips these.
+ *
+ * When we need to see whitespace, we us next_raw_token(), which also
+ * returns the WHITESPACE token.
+ *
+ * The peek_token() and peek_raw_token() functions work as expected.
+ *
+ * Warning: if you invoke peek_token(), then if there is a whitespace
+ * token, it will be lost, and subsequent use of next_raw_token() or
+ * peek_raw_token() will NOT see it.
+ */
+
+static enum dhcp_token
+get_raw_token(struct parse *cfile) {
 	int c;
 	enum dhcp_token ttok;
 	static char tb [2];
-	int l, p, u;
+	int l, p;
 
 	do {
 		l = cfile -> line;
 		p = cfile -> lpos;
-		u = cfile -> ugflag;
 
 		c = get_char (cfile);
-#ifdef OLD_LEXER
-		if (c == '\n' && p == 1 && !u
-		    && cfile -> comment_index < sizeof cfile -> comments)
-			cfile -> comments [cfile -> comment_index++] = '\n';
-#endif
-
-		if (!(c == '\n' && cfile -> eol_token)
-		    && isascii (c) && isspace (c))
-			continue;
+		if (!((c == '\n') && cfile->eol_token) && 
+		    isascii(c) && isspace(c)) {
+		    	ttok = read_whitespace(c, cfile);
+			break;
+		}
 		if (c == '#') {
-#ifdef OLD_LEXER
-			if (cfile -> comment_index < sizeof cfile -> comments)
-			    cfile -> comments [cfile -> comment_index++] = '#';
-#endif
 			skip_to_eol (cfile);
 			continue;
 		}
@@ -223,11 +290,17 @@ static enum dhcp_token get_token (cfile)
 	return ttok;
 }
 
-enum dhcp_token next_token (rval, rlen, cfile)
-	const char **rval;
-	unsigned *rlen;
-	struct parse *cfile;
-{
+/*
+ * The get_next_token() function consumes the next token and
+ * returns it to the caller.
+ *
+ * Since the code is almost the same for "normal" and "raw" 
+ * input, we pass a flag to alter the way it works.
+ */
+
+static enum dhcp_token 
+get_next_token(const char **rval, unsigned *rlen, 
+	       struct parse *cfile, isc_boolean_t raw) {
 	int rv;
 
 	if (cfile -> token) {
@@ -238,9 +311,17 @@ enum dhcp_token next_token (rval, rlen, cfile)
 		rv = cfile -> token;
 		cfile -> token = 0;
 	} else {
-		rv = get_token (cfile);
+		rv = get_raw_token(cfile);
 		cfile -> token_line = cfile -> cur_line;
 	}
+
+	if (!raw) {
+		while (rv == WHITESPACE) {
+			rv = get_raw_token(cfile);
+			cfile->token_line = cfile->cur_line;
+		}
+	}
+	
 	if (rval)
 		*rval = cfile -> tval;
 	if (rlen)
@@ -251,17 +332,56 @@ enum dhcp_token next_token (rval, rlen, cfile)
 	return rv;
 }
 
-enum dhcp_token peek_token (rval, rlen, cfile)
-	const char **rval;
-	unsigned int *rlen;
-	struct parse *cfile;
-{
+
+/*
+ * Get the next token from cfile and return it.
+ *
+ * If rval is non-NULL, set the pointer it contains to 
+ * the contents of the token.
+ *
+ * If rlen is non-NULL, set the integer it contains to 
+ * the length of the token.
+ */
+
+enum dhcp_token
+next_token(const char **rval, unsigned *rlen, struct parse *cfile) {
+	return get_next_token(rval, rlen, cfile, ISC_FALSE);
+}
+
+
+/*
+ * The same as the next_token() function above, but will return space
+ * as the WHITESPACE token.
+ */
+
+enum dhcp_token
+next_raw_token(const char **rval, unsigned *rlen, struct parse *cfile) {
+	return get_next_token(rval, rlen, cfile, ISC_TRUE);
+}
+
+
+/*
+ * The do_peek_token() function checks the next token without
+ * consuming it, and returns it to the caller.
+ *
+ * Since the code is almost the same for "normal" and "raw" 
+ * input, we pass a flag to alter the way it works. (See the 
+ * warning in the GENERAL NOTES ABOUT TOKENS above though.)
+ */
+
+enum dhcp_token
+do_peek_token(const char **rval, unsigned int *rlen,
+	      struct parse *cfile, isc_boolean_t raw) {
 	int x;
 
-	if (!cfile -> token) {
+	if (!cfile->token || (!raw && (cfile->token == WHITESPACE))) {
 		cfile -> tlpos = cfile -> lexchar;
 		cfile -> tline = cfile -> lexline;
-		cfile -> token = get_token (cfile);
+
+		do {
+			cfile->token = get_raw_token(cfile);
+		} while (!raw && (cfile->token == WHITESPACE));
+
 		if (cfile -> lexline != cfile -> tline)
 			cfile -> token_line = cfile -> prev_line;
 
@@ -283,6 +403,36 @@ enum dhcp_token peek_token (rval, rlen, cfile)
 	return cfile -> token;
 }
 
+
+/*
+ * Get the next token from cfile and return it, leaving it for a 
+ * subsequent call to next_token().
+ *
+ * Note that it WILL consume whitespace tokens.
+ *
+ * If rval is non-NULL, set the pointer it contains to 
+ * the contents of the token.
+ *
+ * If rlen is non-NULL, set the integer it contains to 
+ * the length of the token.
+ */
+
+enum dhcp_token
+peek_token(const char **rval, unsigned *rlen, struct parse *cfile) {
+	return do_peek_token(rval, rlen, cfile, ISC_FALSE);
+}
+
+
+/*
+ * The same as the peek_token() function above, but will return space
+ * as the WHITESPACE token.
+ */
+
+enum dhcp_token
+peek_raw_token(const char **rval, unsigned *rlen, struct parse *cfile) {
+	return do_peek_token(rval, rlen, cfile, ISC_TRUE);
+}
+
 static void skip_to_eol (cfile)
 	struct parse *cfile;
 {
@@ -291,16 +441,40 @@ static void skip_to_eol (cfile)
 		c = get_char (cfile);
 		if (c == EOF)
 			return;
-#ifdef OLD_LEXER
-		if (cfile -> comment_index < sizeof (cfile -> comments))
-			comments [cfile -> comment_index++] = c;
-#endif
 		if (c == EOL) {
 			return;
 		}
 	} while (1);
 }
 
+static enum dhcp_token
+read_whitespace(int c, struct parse *cfile) {
+	int ofs;
+
+	/*
+	 * Read as much whitespace as we have available.
+	 */
+	ofs = 0;
+	do {
+		cfile->tokbuf[ofs++] = c;
+		c = get_char(cfile);
+	} while (!((c == '\n') && cfile->eol_token) && 
+		 isascii(c) && isspace(c));
+
+	/*
+	 * Put the last (non-whitespace) character back.
+	 */
+	unget_char(cfile, c);
+
+	/*
+	 * Return our token.
+	 */
+	cfile->tokbuf[ofs] = '\0';
+	cfile->tlen = ofs;
+	cfile->tval = cfile->tokbuf;
+	return WHITESPACE;
+}
+
 static enum dhcp_token read_string (cfile)
 	struct parse *cfile;
 {
@@ -414,9 +588,6 @@ static enum dhcp_token read_number (c, cfile)
 	int c;
 	struct parse *cfile;
 {
-#ifdef OLD_LEXER
-	int seenx = 0;
-#endif
 	int i = 0;
 	int token = NUMBER;
 
@@ -424,7 +595,6 @@ static enum dhcp_token read_number (c, cfile)
 	for (; i < sizeof cfile -> tokbuf; i++) {
 		c = get_char (cfile);
 
-#ifndef OLD_LEXER
 		/* Promote NUMBER -> NUMBER_OR_NAME -> NAME, never demote.
 		 * Except in the case of '0x' syntax hex, which gets called
 		 * a NAME at '0x', and returned to NUMBER_OR_NAME once it's
@@ -458,26 +628,12 @@ static enum dhcp_token read_number (c, cfile)
 			 * token.  If not EOF, rewind the file one byte so
 			 * the next token is read from there.
 			 */
-			if(c != EOF) {
-				cfile->bufix--;
-				cfile->ugflag = 1;
-			}
+			unget_char(cfile, c);
 			goto end_read;
 
 		    default:
 			log_fatal("read_number():%s:%d: impossible case", MDL);
 		}
-#else /* OLD_LEXER */
-		if (!seenx && (c == 'x')) {
-			seenx = 1;
-		} else if (!isascii (c) || !isxdigit (c)) {
-			if (c != EOF) {
-				cfile -> bufix--;
-				cfile -> ugflag = 1;
-			}
-			break;
-		}
-#endif /* OLD_LEXER */
 
 		cfile -> tokbuf [i] = c;
 	}
@@ -507,10 +663,7 @@ static enum dhcp_token read_num_or_name (c, cfile)
 		c = get_char (cfile);
 		if (!isascii (c) ||
 		    (c != '-' && c != '_' && !isalnum (c))) {
-			if (c != EOF) {
-				cfile -> bufix--;
-				cfile -> ugflag = 1;
-			}
+		    	unget_char(cfile, c);
 			break;
 		}
 		if (!isxdigit (c))
@@ -524,17 +677,15 @@ static enum dhcp_token read_num_or_name (c, cfile)
 	cfile -> tokbuf [i] = 0;
 	cfile -> tlen = i;
 	cfile -> tval = cfile -> tokbuf;
-	return intern (cfile -> tval, rv);
+	return intern(cfile->tval, rv);
 }
 
-static enum dhcp_token intern (atom, dfv)
-	unsigned char *atom;
-	enum dhcp_token dfv;
-{
-	if (!isascii (atom [0]))
+static enum dhcp_token
+intern(char *atom, enum dhcp_token dfv) {
+	if (!isascii(atom[0]))
 		return dfv;
 
-	switch (tolower (atom [0])) {
+	switch (tolower((unsigned char)atom[0])) {
 	      case '-':
 		if (atom [1] == 0)
 			return MINUS;
@@ -563,6 +714,8 @@ static enum dhcp_token intern (atom, dfv)
 			return ALIAS;
 		if (!strcasecmp (atom + 1, "lgorithm"))
 			return ALGORITHM;
+		if (!strcasecmp (atom + 1, "lso"))
+			return TOKEN_ALSO;
 		if (!strcasecmp (atom + 1, "bandoned"))
 			return TOKEN_ABANDONED;
 		if (!strcasecmp (atom + 1, "dd"))
@@ -579,6 +732,8 @@ static enum dhcp_token intern (atom, dfv)
 			return TOKEN_ACTIVE;
 		if (!strcasecmp (atom + 1, "tsfp"))
 			return ATSFP;
+                if (!strcasecmp (atom + 1, "fter"))
+                        return AFTER;
 		break;
 	      case 'b':
 		if (!strcasecmp (atom + 1, "ackup"))
@@ -607,68 +762,47 @@ static enum dhcp_token intern (atom, dfv)
 			return BOUND;
 		break;
 	      case 'c':
-		if (!strcasecmp(atom + 1, "ase"))
+		if (!strcasecmp (atom + 1, "ase"))
 			return CASE;
-		if (!strcasecmp(atom + 1, "heck"))
+		if (!strcasecmp (atom + 1, "ommit"))
+			return COMMIT;
+		if (!strcasecmp (atom + 1, "ode"))
+			return CODE;
+		if (!strcasecmp (atom + 1, "onfig-option"))
+			return CONFIG_OPTION;
+		if (!strcasecmp (atom + 1, "heck"))
 			return CHECK;
-		if (!strcasecmp(atom + 1, "iaddr"))
-			return CIADDR;
-		if (isascii(atom[1]) &&
-		    tolower((unsigned char)atom[1]) == 'l') {
-			if (!strcasecmp(atom + 2, "ass"))
-				return CLASS;
-			if (!strncasecmp(atom + 2, "ient", 4)) {
-				if (!strcasecmp(atom + 6, "s"))
-					return CLIENTS;
-				if (atom[6] == '-') {
-					if (!strcasecmp(atom + 7, "hostname"))
-						return CLIENT_HOSTNAME;
-					if (!strcasecmp(atom + 7, "identifier"))
-						return CLIENT_IDENTIFIER;
-					if (!strcasecmp(atom + 7, "state"))
-						return CLIENT_STATE;
-					if (!strcasecmp(atom + 7, "updates"))
-						return CLIENT_UPDATES;
-					break;
-				}
-				break;
-			}
-			if (!strcasecmp(atom + 2, "ose"))
-				return TOKEN_CLOSE;
-			if (!strcasecmp(atom + 2, "tt"))
-				return CLTT;
-			break;
-		}
-		if (isascii(atom[1]) &&
-		    tolower((unsigned char)atom[1]) == 'o') {
-			if (!strcasecmp(atom + 2, "de"))
-				return CODE;
-			if (isascii(atom[2]) &&
-			    tolower((unsigned char)atom[2]) == 'm') {
-				if (!strcasecmp(atom + 3, "mit"))
-					return COMMIT;
-				if (!strcasecmp(atom + 3,
-						"munications-interrupted"))
-					return COMMUNICATIONS_INTERRUPTED;
-				break;
-			}
-			if (isascii(atom[2]) &&
-			    tolower((unsigned char)atom[2]) == 'n') {
-				if (!strcasecmp(atom + 3, "cat"))
-					return CONCAT;
-				if (!strcasecmp(atom + 3, "fig-option"))
-					return CONFIG_OPTION;
-				if (!strcasecmp(atom + 3, "flict-done"))
-					return CONFLICT_DONE;
-				if (!strcasecmp(atom + 3, "nect"))
-					return CONNECT;
-				break;
-			}
-			break;
-		}
-		if (!strcasecmp(atom + 1, "reate"))
+		if (!strcasecmp (atom + 1, "lass"))
+			return CLASS;
+		if (!strcasecmp (atom + 1, "lose"))
+			return TOKEN_CLOSE;
+		if (!strcasecmp(atom + 1, "ompressed"))
+			return COMPRESSED;
+		if (!strcasecmp (atom + 1, "reate"))
 			return TOKEN_CREATE;
-  		break;
+		if (!strcasecmp (atom + 1, "iaddr"))
+			return CIADDR;
+		if (!strncasecmp (atom + 1, "lient", 5)) {
+			if (!strcasecmp (atom + 6, "-identifier"))
+				return CLIENT_IDENTIFIER;
+			if (!strcasecmp (atom + 6, "-hostname"))
+				return CLIENT_HOSTNAME;
+			if (!strcasecmp (atom + 6, "-state"))
+				return CLIENT_STATE;
+			if (!strcasecmp (atom + 6, "-updates"))
+				return CLIENT_UPDATES;
+			if (!strcasecmp (atom + 6, "s"))
+				return CLIENTS;
+		}
+		if (!strcasecmp (atom + 1, "oncat"))
+			return CONCAT;
+		if (!strcasecmp (atom + 1, "onnect"))
+			return CONNECT;
+		if (!strcasecmp (atom + 1, "ommunications-interrupted"))
+			return COMMUNICATIONS_INTERRUPTED;
+		if (!strcasecmp (atom + 1, "ltt"))
+			return CLTT;
+		break;
 	      case 'd':
 		if (!strcasecmp(atom + 1, "b-time-format"))
 			return DB_TIME_FORMAT;
@@ -697,6 +831,8 @@ static enum dhcp_token intern (atom, dfv)
 		if (!strncasecmp (atom + 1, "efault", 6)) {
 			if (!atom [7])
 				return DEFAULT;
+			if (!strcasecmp(atom + 7, "-duid"))
+				return DEFAULT_DUID;
 			if (!strcasecmp (atom + 7, "-lease-time"))
 				return DEFAULT_LEASE_TIME;
 			break;
@@ -726,7 +862,8 @@ static enum dhcp_token intern (atom, dfv)
 		}
 		break;
 	      case 'e':
-		if (isascii (atom [1]) && tolower (atom [1]) == 'x') {
+		if (isascii (atom [1]) && 
+		    tolower((unsigned char)atom[1]) == 'x') {
 			if (!strcasecmp (atom + 2, "tract-int"))
 				return EXTRACT_INT;
 			if (!strcasecmp (atom + 2, "ists"))
@@ -761,6 +898,9 @@ static enum dhcp_token intern (atom, dfv)
 			return ENCAPSULATE;
 		if (!strcasecmp(atom + 1, "xecute"))
 			return EXECUTE;
+		if (!strcasecmp(atom+1, "n")) {
+			return EN;
+		}
 		break;
 	      case 'f':
 		if (!strcasecmp (atom + 1, "atal"))
@@ -769,6 +909,8 @@ static enum dhcp_token intern (atom, dfv)
 			return FILENAME;
 		if (!strcasecmp (atom + 1, "ixed-address"))
 			return FIXED_ADDR;
+		if (!strcasecmp (atom + 1, "ixed-address6"))
+			return FIXED_ADDR6;
 		if (!strcasecmp (atom + 1, "ddi"))
 			return FDDI;
 		if (!strcasecmp (atom + 1, "ormerr"))
@@ -797,6 +939,8 @@ static enum dhcp_token intern (atom, dfv)
 			return HOST;
 		if (!strcasecmp (atom + 1, "ost-decl-name"))
 			return HOST_DECL_NAME;
+		if (!strcasecmp(atom + 1, "ost-identifier"))
+			return HOST_IDENTIFIER;
 		if (!strcasecmp (atom + 1, "ardware"))
 			return HARDWARE;
 		if (!strcasecmp (atom + 1, "ostname"))
@@ -805,6 +949,10 @@ static enum dhcp_token intern (atom, dfv)
 			return TOKEN_HELP;
 		break;
 	      case 'i':
+	      	if (!strcasecmp(atom+1, "a-na")) 
+			return IA_NA;
+	      	if (!strcasecmp(atom+1, "aaddr")) 
+			return IAADDR;
 		if (!strcasecmp (atom + 1, "nclude"))
 			return INCLUDE;
 		if (!strcasecmp (atom + 1, "nteger"))
@@ -815,6 +963,8 @@ static enum dhcp_token intern (atom, dfv)
 			return INFO;
 		if (!strcasecmp (atom + 1, "p-address"))
 			return IP_ADDRESS;
+		if (!strcasecmp (atom + 1, "p6-address"))
+			return IP6_ADDRESS;
 		if (!strcasecmp (atom + 1, "nitial-interval"))
 			return INITIAL_INTERVAL;
 		if (!strcasecmp (atom + 1, "nterface"))
@@ -844,6 +994,8 @@ static enum dhcp_token intern (atom, dfv)
 			return LCASE;
 		if (!strcasecmp (atom + 1, "ease"))
 			return LEASE;
+		if (!strcasecmp(atom + 1, "ease6"))
+			return LEASE6;
 		if (!strcasecmp (atom + 1, "eased-address"))
 			return LEASED_ADDRESS;
 		if (!strcasecmp (atom + 1, "ease-time"))
@@ -862,6 +1014,12 @@ static enum dhcp_token intern (atom, dfv)
 			return LOCAL;
 		if (!strcasecmp (atom + 1, "og"))
 			return LOG;
+		if (!strcasecmp(atom+1, "lt")) {
+			return LLT;
+		}
+		if (!strcasecmp(atom+1, "l")) {
+			return LL;
+		}
 		break;
 	      case 'm':
 		if (!strncasecmp (atom + 1, "ax", 2)) {
@@ -877,6 +1035,8 @@ static enum dhcp_token intern (atom, dfv)
 				if (!strcasecmp(atom + 10, "time"))
 					return MAX_LEASE_TIME;
 			}
+			if (!strcasecmp(atom + 3, "-life"))
+				return MAX_LIFE;
 			if (!strcasecmp (atom + 3, "-transmit-idle"))
 				return MAX_TRANSMIT_IDLE;
 			if (!strcasecmp (atom + 3, "-response-delay"))
@@ -923,7 +1083,7 @@ static enum dhcp_token intern (atom, dfv)
 		if (!strcasecmp (atom + 1, "ot"))
 			return TOKEN_NOT;
 		if (!strcasecmp (atom + 1, "o"))
-			return NO;
+			return TOKEN_NO;
 		if (!strcasecmp (atom + 1, "s-update"))
 			return NS_UPDATE;
 		if (!strcasecmp (atom + 1, "oerror"))
@@ -966,6 +1126,8 @@ static enum dhcp_token intern (atom, dfv)
 	      case 'p':
 		if (!strcasecmp (atom + 1, "repend"))
 			return PREPEND;
+		if (!strcasecmp(atom + 1, "referred-life"))
+			return PREFERRED_LIFE;
 		if (!strcasecmp (atom + 1, "acket"))
 			return PACKET;
 		if (!strcasecmp (atom + 1, "ool"))
@@ -997,6 +1159,9 @@ static enum dhcp_token intern (atom, dfv)
 			return RESOLUTION_INTERRUPTED;
 		if (!strcasecmp (atom + 1, "ange"))
 			return RANGE;
+		if (!strcasecmp(atom + 1, "ange6")) {
+			return RANGE6;
+		}
 		if (!strcasecmp (atom + 1, "ecover"))
 			return RECOVER;
 		if (!strcasecmp (atom + 1, "ecover-done"))
@@ -1043,10 +1208,12 @@ static enum dhcp_token intern (atom, dfv)
 	      case 's':
                 if (!strcasecmp(atom + 1, "cript"))
                         return SCRIPT;
-		if (isascii(atom[1]) && tolower(atom[1]) == 'e') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 'e') {
                         if (!strcasecmp(atom + 2, "arch"))
                                 return SEARCH;
-			if (isascii(atom[2]) && tolower(atom[2]) == 'c') {
+			if (isascii(atom[2]) && 
+			    tolower((unsigned char)atom[2]) == 'c') {
 				if (!strncasecmp(atom + 3, "ond", 3)) {
                                         if (!strcasecmp(atom + 6, "ary"))
                                                 return SECONDARY;
@@ -1070,8 +1237,11 @@ static enum dhcp_token intern (atom, dfv)
 			if (!strncasecmp(atom + 2, "rv", 2)) {
 				if (!strncasecmp(atom + 4, "er", 2)) {
                                         if (atom[6] == '\0')
-                                                return SERVER;
+                                                return TOKEN_SERVER;
 					if (atom[6] == '-') {
+						if (!strcasecmp(atom + 7,
+								"duid")) 
+							return SERVER_DUID;
                                                 if (!strcasecmp(atom + 7,
 								"name"))
                                                         return SERVER_NAME;
@@ -1090,14 +1260,16 @@ static enum dhcp_token intern (atom, dfv)
                                 return TOKEN_SET;
 			break;
 		}
-		if (isascii(atom[1]) && tolower(atom[1]) == 'h') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 'h') {
                         if (!strcasecmp(atom + 2, "ared-network"))
                                 return SHARED_NETWORK;
                         if (!strcasecmp(atom + 2, "utdown"))
                                 return SHUTDOWN;
 			break;
 		}
-		if (isascii(atom[1]) && tolower(atom[1]) == 'i') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 'i') {
                         if (!strcasecmp(atom + 2, "addr"))
                                 return SIADDR;
                         if (!strcasecmp(atom + 2, "gned"))
@@ -1106,8 +1278,10 @@ static enum dhcp_token intern (atom, dfv)
                                 return SIZE;
 			break;
 		}
-		if (isascii(atom[1]) && tolower(atom[1]) == 'p') {
-			if (isascii(atom[2]) && tolower(atom[2]) == 'a') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 'p') {
+			if (isascii(atom[2]) && 
+			    tolower((unsigned char)atom[2]) == 'a') {
                                 if (!strcasecmp(atom + 3, "ce"))
                                         return SPACE;
                                 if (!strcasecmp(atom + 3, "wn"))
@@ -1118,8 +1292,10 @@ static enum dhcp_token intern (atom, dfv)
                                 return SPLIT;
 			break;
 		}
-		if (isascii(atom[1]) && tolower(atom[1]) == 't') {
-			if (isascii(atom[2]) && tolower(atom[2]) == 'a') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 't') {
+			if (isascii(atom[2]) && 
+			    tolower((unsigned char)atom[2]) == 'a') {
 				if(!strncasecmp(atom + 3, "rt", 2)) {
                                          if (!strcasecmp(atom + 5, "s"))
                                                  return STARTS;
@@ -1128,7 +1304,7 @@ static enum dhcp_token intern (atom, dfv)
 					break;
 				}
 				if (isascii(atom[3]) &&
-				    tolower(atom[3]) == 't') {
+				    tolower((unsigned char)atom[3]) == 't') {
                                         if (!strcasecmp(atom + 4, "e"))
                                                 return STATE;
                                         if (!strcasecmp(atom + 4, "ic"))
@@ -1145,11 +1321,14 @@ static enum dhcp_token intern (atom, dfv)
                                 return SUBCLASS;
                         if (!strcasecmp(atom + 3, "net"))
                                 return SUBNET;
+                        if (!strcasecmp(atom + 3, "net6"))
+                                return SUBNET6;
                         if (!strcasecmp(atom + 3, "string"))
                                 return SUBSTRING;
                         break;
                 }
-		if (isascii(atom[1]) && tolower(atom[1]) == 'u') {
+		if (isascii(atom[1]) && 
+		    tolower((unsigned char)atom[1]) == 'u') {
                         if (!strcasecmp(atom + 2, "ffix"))
                                 return SUFFIX;
                         if (!strcasecmp(atom + 2, "persede"))
